@@ -6,4 +6,5 @@ eval "$(./conda/bin/conda shell.bash hook)"
 conda activate ./env
 
 python -m torch.utils.collect_env
-pytest --cov=torchtext --junitxml=test-results/junit.xml -v --durations 20 test
+cd test
+pytest --cov=torchtext --junitxml=test-results/junit.xml -v --durations 20 torchtext_unittest
