@@ -1,6 +1,7 @@
 #!/usr/bin/env bash
 
 unset PYTORCH_VERSION
+unset TORCHDATA_VERSION
 # For unittest, nightly PyTorch is used as the following section,
 # so no need to set PYTORCH_VERSION.
 # In fact, keeping PYTORCH_VERSION forces us to hardcode PyTorch version in config.
@@ -18,7 +19,8 @@ conda activate ./env
 printf "* Installing PyTorch\n"
 conda install -y -c "pytorch-${UPLOAD_CHANNEL}" ${CONDA_CHANNEL_FLAGS} pytorch cpuonly
 
-printf "* Installing torchdata nightly\n"
+printf "* Installing torchdata nightly with portalocker\n"
+pip install "portalocker>=2.0.0"
 pip install --pre torchdata --extra-index-url https://download.pytorch.org/whl/nightly/cpu
 
 printf "* Installing pywin32_postinstall script\n"
