@@ -1,6 +1,7 @@
 from .bundler import (
     ROBERTA_BASE_ENCODER,
     ROBERTA_LARGE_ENCODER,
+    ROBERTA_DISTILLED_ENCODER,
     RobertaBundle,
     XLMR_BASE_ENCODER,
     XLMR_LARGE_ENCODER,
@@ -16,4 +17,5 @@ __all__ = [
     "XLMR_LARGE_ENCODER",
     "ROBERTA_BASE_ENCODER",
     "ROBERTA_LARGE_ENCODER",
+    "ROBERTA_DISTILLED_ENCODER",
 ]
