@@ -47,10 +47,8 @@ def validate_file(file_obj, hash_value, hash_type="sha256"):
         bool: return True if its a valid file, else False.
     """
 
-    if hash_type == "sha256":
-        hash_func = hashlib.sha256()
-    elif hash_type == "md5":
-        hash_func = hashlib.md5()
+    if hash_type in ("sha256", "md5"):
+        hash_func = hashlib.new(hash_type, usedforsecurity=False)
     else:
         raise ValueError
 
@@ -209,8 +207,8 @@ def _log_class_usage(klass):
     torch._C._log_api_usage_once(identifier)
 
 
-def get_asset_local_path(asset_path: str, overwite=False) -> str:
-    """Get local path for assets. Download if path does not exost locally
+def get_asset_local_path(asset_path: str, overwrite=False) -> str:
+    """Get local path for assets. Download if path does not exist locally
     Args:
         asset_path: Local path to asset or remote URL
         overwrite: Indicate whether to overwrite the file when downloading from URL (default: False)
@@ -226,5 +224,5 @@ def get_asset_local_path(asset_path: str, overwite=False) -> str:
     if os.path.exists(asset_path):
         local_path = asset_path
     else:
-        local_path = download_from_url(url=asset_path, root=_CACHE_DIR, overwrite=overwite)
+        local_path = download_from_url(url=asset_path, root=_CACHE_DIR, overwrite=overwrite)
     return local_path
